#!/usr/bin/env python

import os
import win32process
import sys
import atexit
import socket
import time
import re
import signal

LOCKFILE=".eclient.lock"
CRLF="\r\n"

class Eclient:
    """This class manages a connection to the Enigma Key Server."""

    def __init__(self, host=None, port=0):
        """Constructor. When called without arguments,
           create an unconnected instance.
        """
        self.host = host
        self.port = port
        self.sock = None
        self.file = None

    def usage(self):
        """Display usage message"""
        print "Usage: enigma-client host port"
        sys.exit(1)
    
    def connect(self, host=None, port=0):
        """ Create socket and connect to a host on a given port."""
        self.sock = None
        self.host = host
        self.port = port
        try:
            for res in socket.getaddrinfo(host, port, 0, socket.SOCK_STREAM):
                af, socktype, proto, canonname, sa = res
                try:
                    self.sock = socket.socket(af, socktype, proto)
                    self.sock.connect(sa)
                except socket.error:
                    if self.sock:
                        self.sock.close()
                    self.sock = None
                    continue
                break
        except socket.gaierror:
            return

    def persistent_connect(self, host=None, port=0, interval=1800):
        self.sock = None
        while not self.sock:
            self.connect(host, port);
            if not self.sock:
                time.sleep(interval)

    def close(self):
        """Close the connection to the enigma key server."""
        if self.file:
            self.file.close()
        self.file = None
        if self.sock:
            self.sock.close()
        self.sock = None

    def sendline(self, str):
        """Send 'str' postfixed with CRLF to the server."""
        str = str.rstrip()
        str = '%s%s' % (str, CRLF)
        try:
            self.sock.sendall(str)
        except socket.error:
            self.close()
            return 0
        return 1
 
    def getline(self):
        """Attempts to read a single line from the network buffer.
           Replaces CRLF with \n. Returns the line or 'None' on Error.
        """
        if self.file is None:
            self.file = self.sock.makefile('rb')
        try:
            line = self.file.readline()
        except socket.error:
            self.close()
            return None
        if line == '':
            self.close()
            return None
        line = re.sub(CRLF, "\n", line)
        return line
    
    def filewrite(self, file, str):
        """Writes 'str' to file."""
        try:
            f = open(file, 'w')
        except EnvironmentError:
            return 0
        try:
            f.write(str)
        except EnvironmentError:
            f.close()
            return 0
        f.close()
        return 1

    def fileread(self, file):
        """Returns contents of file in a string"""
        try:
            f = open(file, 'r')
        except EnvironmentError:
            return None
        try:
            s = f.read()
        except EnvironmentError:
            f.close()
            return None
        f.close()
        return s

    def filereadlines(self, file):
        """Returns contents of file in a list of lines"""
        try:
            f = open(file, 'r')
        except EnvironmentError:
            return None
        try:
            list = f.readlines()
        except EnvironmentError:
            f.close()
            return None
        f.close()
        return list

    def get_chunk(self, host, port):
        """Attempts to get a chunk for ./enigma to work on.
           Retries until successful.
        """
        done = 0
        while not done:
            self.persistent_connect(host, port, 600)
            if not self.sendline('WANT_CHUNK'):
                time.sleep(10)
                continue
            reply = self.getline()        
            if reply == 'TRY_AGAIN_LATER\n':
                self.close()
                time.sleep(1800)
                continue
            elif reply == 'OK\n':
                i = 0
                c = 0
                data = [None]*4
                while i < 4:
                    data[i] = self.getline()
                    if data[i] is None:
                        c = 1
                        break
                    i += 1
                if c:
                    self.close()
                    time.sleep(10)
                    continue
                ticket = data[0]
                chunk = ''.join([data[1], data[2]])
                ciphertext = data[3]
    
                self.close()
                while not self.filewrite('00ciphertext', ciphertext):
                    time.sleep(60)
                while not self.filewrite('00ticket', ticket):
                    time.sleep(60)
                while not self.filewrite('00hc.resume', chunk):
                    time.sleep(60)
                done = 1
            else:
                self.close()
                time.sleep(10)
                continue
    
    def submit_chunk(self, host, port):
        """Attempts to submit a chunk that has been processed by ./enigma
           Retries until successful.
        """
        ticket = self.fileread('00ticket')
        if ticket is None:
            return
        chunk = self.filereadlines('00hc.resume')
        if chunk is None:
            return
        done = 0
        while not done:
            self.persistent_connect(host, port, 600)
            if not self.sendline('HAVE_CHUNK'):
                time.sleep(10)
                continue
            reply = self.getline()        
            if reply != 'SEND_TICKET\n':
                self.close()
                time.sleep(10)
                continue
            if not self.sendline(ticket):
                time.sleep(10)
                continue
            reply = self.getline()
            if reply != 'SEND_CHUNK\n':
                self.close()
                time.sleep(10)
                continue;
            if not self.sendline(chunk[0]):
                time.sleep(10)
                continue
            if not self.sendline(chunk[1]):
                time.sleep(10)
                continue
            reply = self.getline()
            if reply == 'TRY_AGAIN_LATER\n':
                self.close()
                time.sleep(1800)
                continue
            else:
                self.close()
                done = 1

    def touch_lockfile(self):
        lockfile = open(LOCKFILE, 'w')
        lockfile.close()

    def rm_lockfile(self):
        os.remove(LOCKFILE)


""" main """
cmdline = 'enigma.exe -R 00trigr.naval 00bigr.naval 00ciphertext > NUL'
eclient = Eclient()

if len(sys.argv) != 3:
    eclient.usage()

if os.path.isfile(LOCKFILE):
    print "enigma-client: error: found lockfile %s. \n" \
          "Check that no other enigma-client process is using this directory." \
          % LOCKFILE
    sys.exit(1)

atexit.register(eclient.rm_lockfile)
eclient.touch_lockfile()
# WINDOWS
win32process.SetPriorityClass(
   win32process.GetCurrentProcess(),
   win32process.IDLE_PRIORITY_CLASS
)
# signal.signal(signal.SIGPIPE, signal.SIG_IGN)
# End WINDOWS

while 1:
    retval =  os.system(cmdline) >> 8
    if retval == 0:
        eclient.submit_chunk(sys.argv[1], int(sys.argv[2]))
        eclient.get_chunk(sys.argv[1], int(sys.argv[2]))
    elif retval == 1:
        eclient.get_chunk(sys.argv[1], int(sys.argv[2]))
        time.sleep(10)
    else:
        """./enigma has caught a signal"""
        sys.exit(retval)


#
# This file is part of enigma-suite-0.72, which is distributed under the terms
# of the General Public License (GPL), version 2. See doc/COPYING for details.
#
# Copyright (C) 2005 Stefan Krah
# 
